#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#import <sndfile.h>

#define BUFFER_SIZE 32768

int main(int argc, char *argv[])
{
	int len,i,n,nindex=0,ntrack=1,ntrack_all=0,block,size[99],type[99],stat=1,bps,blockper,received,lastFile,lastDecode,bytesToReceive,blank;
	char cuebuf[256],outfile[512],ext[10],*buffer;
	FILE *fp;
	SF_INFO sfinfo;
	SNDFILE *snd_data, *snd_data_w;
	
	if(argc != 3) {
		fprintf(stderr, "usage: %s <cue sheet> <file to split>\n",argv[0]);
		return 0;
	}
	
	if((fp = fopen(argv[1], "rb")) == NULL) {
		fprintf(stderr,"error: cannot open cue sheet\n");
		return 0;
	}
	
	snd_data = sf_open(argv[2], SFM_READ, &sfinfo);
	if(sf_error(snd_data)) {
		fprintf(stderr,"error: cannot open file %s\n",argv[2]);
		return 0;
	}
	switch((sfinfo.format)&SF_FORMAT_SUBMASK) {
	  case 1:
		bps = 1;
		break;
	  case 2:
		bps = 2;
		break;
	  case 3:
		bps = 3;
		break;
	  case 4:
		bps = 4;
		break;
	  case 5:
		bps = 1;
		break;
	  case 6:
		bps = 4;
		break;
	  case 7:
		bps = 8;
		break;
	  default:
		fprintf(stderr,"error: unknown format");
		return 0;
	}
	
	switch((sfinfo.format)&SF_FORMAT_TYPEMASK) {
	  case 0x10000:
		strcpy(ext,"wav");
		break;
	  case 0x20000:
		strcpy(ext,"aif");
		break;
	  case 0x30000:
		strcpy(ext,"au");
		break;
	  default:
		ext[0] = '\0';
	}
	
	while(fgets(cuebuf,256,fp) != NULL) {
		len = strlen(cuebuf);
		if(stat == 1) {
			i=0;
			while(*(cuebuf+i)==' ' || *(cuebuf+i)=='\t') i++;
			if(i>len-4) continue;
			if(!strncmp(cuebuf+i,"TRACK",5)) {
				stat = 2;
				ntrack_all++;
			}
		}
		else if(stat == 2) {
			i=0;
			while(*(cuebuf+i)==' ' || *(cuebuf+i)=='\t') i++;
			if(i>len-4) continue;
			if(!strncmp(cuebuf+i,"INDEX",5)) {
				i = i + 5;
				while(*(cuebuf+i)==' ' || *(cuebuf+i)=='\t') i++;
				if(!strncmp(cuebuf+i,"00",2)) {
					type[nindex] = 0;
					i = i + 2;
					while(*(cuebuf+i)==' ' || *(cuebuf+i)=='\t') i++;
					*(cuebuf+i+2) = '\0';
					*(cuebuf+i+5) = '\0';
					*(cuebuf+i+8) = '\0';
					block = atoi(cuebuf+i)*60*sfinfo.samplerate;
					block = block + atoi(cuebuf+i+3)*sfinfo.samplerate;
					block = block + atoi(cuebuf+i+6)*sfinfo.samplerate/100;
					size[nindex++] = block;
				}
				else if(!strncmp(cuebuf+i,"01",2)) {
					type[nindex] = 1;
					i = i + 2;
					while(*(cuebuf+i)==' ' || *(cuebuf+i)=='\t') i++;
					*(cuebuf+i+2) = '\0';
					*(cuebuf+i+5) = '\0';
					*(cuebuf+i+8) = '\0';
					block = atoi(cuebuf+i)*60*sfinfo.samplerate;
					block = block + atoi(cuebuf+i+3)*sfinfo.samplerate;
					block = block + atoi(cuebuf+i+6)*sfinfo.samplerate/100;
					stat = 1;
					size[nindex++] = block;
				}
			}
		}
	}
	fclose(fp);
	
	for(i=nindex-1;i>0;i--) {
		size[i] = size[i] - size[i-1];
	}
	
	for(i=nindex-1;i>0;i--) {
		size[i] = size[i] * bps * sfinfo.channels;
	}
	
	len = strlen(argv[2]);
	for(i=0;i<len;i++) {
		if(*(argv[2]+i) == '.') {
			*(argv[2]+i) = '\0';
		}
	}
	
	if((buffer = malloc(BUFFER_SIZE)) == NULL) {
		sf_close(snd_data);
		fprintf(stderr,"error: malloc error\n");
		return 0;
	}
	
	for(n=0;n<nindex;n++) {
		printf("processing track %d/%d...\r",ntrack,ntrack_all);
		fflush(stdout);
		blockper = BUFFER_SIZE;
		if(n == nindex-1) {
			lastFile = 1;
		}
		else {
			lastFile = 0;
			bytesToReceive = size[n+1];
		}
		if(type[n] == 0) {
			blank = 1;
		}
		else {
			if(ext[0]) sprintf(outfile,"%s.%d.%s",argv[2],ntrack++,ext);
			else sprintf(outfile,"%s.%d",argv[2],ntrack++);
			snd_data_w = sf_open(outfile, SFM_WRITE, &sfinfo);
			if(sf_error(snd_data_w)) {
				free(buffer);
				fprintf(stderr,"error: cannot create file\n");
				return 0;
			}
			blank = 0;
		}
		lastDecode = 0;
		while(1) {
			if(!lastFile && blockper >= bytesToReceive) {
				blockper = bytesToReceive;
				lastDecode = 1;
			}
			
			received = sf_read_raw(snd_data,buffer,blockper);
			
			if(received<0 || sf_error(snd_data)) {
				free(buffer);
				if(!blank) sf_close(snd_data_w);
				sf_close(snd_data);
				fprintf(stderr,"error: cannot read file\n");
				return 0;
			}
			if(!blank) {
				sf_write_raw(snd_data_w, buffer, received);
				if(sf_error(snd_data_w)) {
					free(buffer);
					sf_close(snd_data_w);
					sf_close(snd_data);
					fprintf(stderr,"error: cannot write file\n");
					return 0;
				}
			}
			if(!lastFile) bytesToReceive = bytesToReceive - received;
			if(lastDecode) break;
			if(received<blockper) break;
		}
		if(!blank) sf_close(snd_data_w);
	}
	
	sf_close(snd_data);
	free(buffer);
	printf("\nsuccessful.\n");
	
	return 0;
}